/* array_shortcuts.h -- Definitions of "shortcut" typedefs for array types

    Copyright (C) 2015 European Centre for Medium-Range Weather Forecasts

    Author: Robin Hogan <r.j.hogan@ecmwf.int>

    This file is part of the Adept library.
*/

#ifndef AdeptArrayShortcuts_H
#define AdeptArrayShortcuts_H

#include <adept/Array.h>
#include <adept/SpecialMatrix.h>

namespace adept {

  // ---------------------------------------------------------------------
  // Pretty typedefs to avoid the need for template arguments
  // ---------------------------------------------------------------------

  typedef Array<1> Vector;
  typedef Array<2> Matrix;
  typedef Array<3> Array3;

  typedef Array<1,Index> IntVector;
  typedef Array<2,Index> IntMatrix;
  typedef Array<3,Index> IntArray3;
  typedef Array<1,int> intVector;
  typedef Array<2,int> intMatrix;
  typedef Array<3,int> intArray3;
  typedef Array<1,bool> boolVector;
  typedef Array<2,bool> boolMatrix;
  typedef Array<3,bool> boolArray3;
  typedef Array<1,float> floatVector;
  typedef Array<2,float> floatMatrix;
  typedef Array<3,float> floatArray3;

  typedef SpecialMatrix<Real,internal::SquareEngine<ROW_MAJOR>,
    false> SquareMatrix;
  typedef SpecialMatrix<Real,internal::BandEngine<ROW_MAJOR,0,0>,
    false> DiagMatrix;
  typedef SpecialMatrix<Real,internal::BandEngine<ROW_MAJOR,1,1>,
    false> TridiagMatrix;
  typedef SpecialMatrix<Real,internal::BandEngine<ROW_MAJOR,2,2>,
    false> PentadiagMatrix;
  typedef SpecialMatrix<Real,internal::SymmEngine<ROW_LOWER_COL_UPPER>,
    false> SymmMatrix;
  typedef SpecialMatrix<Real,internal::LowerEngine<ROW_MAJOR>,
    false> LowerMatrix;
  typedef SpecialMatrix<Real,internal::UpperEngine<ROW_MAJOR>,
    false> UpperMatrix;

  // If automatic differentiation is turned off then aVector and
  // friends become identical to their inactive counterparts
#ifdef ADEPT_NO_AUTOMATIC_DIFFERENTIATION
#define ADEPT_IS_ACTIVE false
#else
#define ADEPT_IS_ACTIVE true
#endif

  typedef Array<1,Real,ADEPT_IS_ACTIVE> aVector;
  typedef Array<2,Real,ADEPT_IS_ACTIVE> aMatrix;
  typedef Array<3,Real,ADEPT_IS_ACTIVE> aArray3;

  typedef SpecialMatrix<Real,internal::SquareEngine<ROW_MAJOR>,
    ADEPT_IS_ACTIVE> aSquareMatrix;
  typedef SpecialMatrix<Real,internal::BandEngine<ROW_MAJOR,0,0>,
    ADEPT_IS_ACTIVE> aDiagMatrix;
  typedef SpecialMatrix<Real,internal::BandEngine<ROW_MAJOR,1,1>,
    ADEPT_IS_ACTIVE> aTridiagMatrix;
  typedef SpecialMatrix<Real,internal::BandEngine<ROW_MAJOR,2,2>,
    ADEPT_IS_ACTIVE> aPentadiagMatrix;
  typedef SpecialMatrix<Real,internal::SymmEngine<ROW_LOWER_COL_UPPER>,
    ADEPT_IS_ACTIVE> aSymmMatrix;
  typedef SpecialMatrix<Real,internal::LowerEngine<ROW_MAJOR>,
    ADEPT_IS_ACTIVE> aLowerMatrix;
  typedef SpecialMatrix<Real,internal::UpperEngine<ROW_MAJOR>,
    ADEPT_IS_ACTIVE> aUpperMatrix;

#undef ADEPT_IS_ACTIVE

} // End namespace adept

#endif
